"""Group3 argparse."""
import argparse

__title__ = __doc__


def argument_parser():
    """As required by spicerack._cookbook."""
    return argparse.ArgumentParser("group3.argparse")


def run(_args, _spicerack):
    """As required by spicerack._cookbook."""
